#include <dlfcn.h>
#include <pthread.h>
#include <execinfo.h>
#include <stdlib.h>
#include <libgen.h>
#include <mach/mach_port.h>
#include <libkern/OSAtomic.h>
#import <Cocoa/Cocoa.h>
#import <Carbon/Carbon.h>
#import <IOKit/IOKitLib.h>
#import <objc/Object.h>

//#define BUG854142_CRASH 1

class nsAutoreleasePool {
public:
    nsAutoreleasePool()
    {
        mLocalPool = [[NSAutoreleasePool alloc] init];
    }
    ~nsAutoreleasePool()
    {
        [mLocalPool release];
    }
private:
    NSAutoreleasePool *mLocalPool;
};

char *GetOwnerName(void *address);
char *GetAddressName(void *address);
void PrintAddress(void *address);
void PrintStackTrace(NSArray *addresses = nil);
BOOL SwizzleMethods(Class aClass, SEL orgMethod, SEL posedMethod, BOOL classMethods);

Class gSymbolicatorClass = nil;
static BOOL gSymbolicationLoaded = NO;
static BOOL loadSymbolication()
{
  if (gSymbolicationLoaded) {
    return YES;
  }
  // Needed for printing stack traces.  Available on OS X 10.5 and up.
  if (dlopen("/System/Library/PrivateFrameworks/Symbolication.framework/Symbolication", RTLD_LAZY)) {
    gSymbolicationLoaded = YES;
    Class VMUDebugMapExtractorClass = ::NSClassFromString(@"VMUDebugMapExtractor");
    SwizzleMethods(VMUDebugMapExtractorClass, @selector(debugMapExtractorWithMachOHeader:),
                   @selector(VMUDebugMapExtractor_debugMapExtractorWithMachOHeader:), YES);
    gSymbolicatorClass = ::NSClassFromString(@"VMUSymbolicator");
  }
  return gSymbolicationLoaded;
}

pthread_t gMainThreadID = 0;

bool IsMainThread()
{
  return (gMainThreadID == pthread_self());
}

static int Hooked_pthread_once(pthread_once_t *once_control, void (*init_routine)(void))
{
  int retval = pthread_once(once_control, init_routine);
  if (!gMainThreadID) {
    gMainThreadID = pthread_self();
  }
  return retval;
}

extern "C" void *NSPushAutoreleasePool();

static void *Hooked_NSPushAutoreleasePool()
{
  void *retval = NSPushAutoreleasePool();
  if (IsMainThread()) {
    loadSymbolication();
  }
  return retval;
}

static int32_t sLogging = 0;

#define STACK_CHECK_LIMIT 10

bool IsFromGeForceDriver()
{
  if (sLogging > 0) {
    return false;
  }
  bool retval = false;
  void **addresses = (void **) calloc(STACK_CHECK_LIMIT, sizeof(void *));
  if (addresses) {
    int count = backtrace(addresses, STACK_CHECK_LIMIT);
    for (int i = 2; i < STACK_CHECK_LIMIT; ++i) {
      Dl_info info = {0};
      dladdr(addresses[i], &info);
      char callerName[1024] = {0};
      strncpy(callerName, basename((char *)info.dli_fname),
              sizeof(callerName) - 1);
      if ((callerName && !strcmp(callerName, "libclh.dylib")) ||
          (callerName && !strcmp(callerName, "GeForceGLDriver")))
      {
        retval = true;
        break;
      }
    }
    free(addresses);
  }
  return retval;
}

extern "C" void clhDeviceDestroy(void *arg1);

static void Hooked_clhDeviceDestroy(void *arg1)
{
#if defined(BUG854142_CRASH)
  OSAtomicIncrement32(&sLogging);
  //nsAutoreleasePool localPool;
  NSLog(@"clhDeviceDestroy(): arg1 %p", arg1);
  PrintStackTrace();
  OSAtomicDecrement32(&sLogging);
#endif
  clhDeviceDestroy(arg1);
#if !defined(BUG854142_CRASH)
  OSAtomicIncrement32(&sLogging);
  //nsAutoreleasePool localPool;
  NSLog(@"clhDeviceDestroy(): arg1 %p", arg1);
  PrintStackTrace();
  OSAtomicDecrement32(&sLogging);
#endif
}

extern "C" int32_t clhDeviceGet(void **arg1, void *arg2);

static int32_t sIn_clhDeviceGet = 0;

static int32_t Hooked_clhDeviceGet(void **arg1, void *arg2)
{
  OSAtomicIncrement32(&sLogging);
  //nsAutoreleasePool localPool;
  NSLog(@"clhDeviceGet(1): arg1 %p, *arg1 %p, arg2 %p",
        arg1, arg1 ? *arg1 : nil, arg2);
  OSAtomicDecrement32(&sLogging);
  OSAtomicIncrement32(&sIn_clhDeviceGet);
  int32_t retval = clhDeviceGet(arg1, arg2);
  OSAtomicDecrement32(&sIn_clhDeviceGet);
  OSAtomicIncrement32(&sLogging);
  NSLog(@"clhDeviceGet(2): arg1 %p, *arg1 %p, arg2 %p, returning %i",
        arg1, arg1 ? *arg1 : nil, arg2, retval);
  PrintStackTrace();
  OSAtomicDecrement32(&sLogging);
  return retval;
}

static kern_return_t Hooked_IOConnectMapMemory(io_connect_t connect,
                                               uint32_t memoryType,
                                               task_port_t intoTask,
#if !__LP64__ || defined(IOCONNECT_MAPMEMORY_10_6)
                                               vm_address_t *atAddress,
                                               vm_size_t *ofSize,
#else
                                               mach_vm_address_t *atAddress,
                                               mach_vm_size_t *ofSize,
#endif
                                               IOOptionBits options)
{
  kern_return_t retval = IOConnectMapMemory(connect, memoryType, intoTask,
                                            atAddress, ofSize, options);
  if (sIn_clhDeviceGet > 0) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOConnectMapMemory(): connect %p (userRetainCount %u, kernelRetainCount %u), memoryType %p, intoTask %p, atAddress %p, ofSize %p, options %p, returning %i",
          (void *) connect, IOObjectGetUserRetainCount(connect), IOObjectGetKernelRetainCount(connect),
          (void *) memoryType, (void *) intoTask,
          (void *) (atAddress ? *atAddress : 0),
          (void *) (ofSize ? *ofSize : 0),
          (void *) options, retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
#if defined(BUG854142_CRASH)
    retval = 1;
#endif
  }
  return retval;
}

kern_return_t Hooked_IOConnectCallMethod(mach_port_t connection,
                                         uint32_t selector,
                                         const uint64_t *input,
                                         uint32_t inputCnt,
                                         const void *inputStruct,
                                         size_t inputStructCnt,
                                         uint64_t *output,
                                         uint32_t *outputCnt,
                                         void *outputStruct,
                                         size_t *outputStructCnt)
{
  kern_return_t retval = IOConnectCallMethod(connection, selector, input, inputCnt,
                                             inputStruct, inputStructCnt,
                                             output, outputCnt,
                                             outputStruct,
                                             outputStructCnt);
  if (sIn_clhDeviceGet > 0) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOConnectCallMethod(): connection %p (userRetainCount %u, kernelRetainCount %u), selector %p, returning %i",
         (void *) connection,
         IOObjectGetUserRetainCount(connection), IOObjectGetKernelRetainCount(connection),
         (void *) selector, retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
#if defined(BUG854142_CRASH)
    retval = 1;
#endif
  }
  return retval;
}

static kern_return_t Hooked_IOConnectUnmapMemory(io_connect_t connect,
                                                 uint32_t memoryType,
                                                 task_port_t fromTask,
#if !__LP64__ || defined(IOCONNECT_MAPMEMORY_10_6)
                                                 vm_address_t atAddress)
#else
                                                 mach_vm_address_t atAddress)
#endif
{
  kern_return_t retval = IOConnectUnmapMemory(connect, memoryType,
                                              fromTask, atAddress);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOConnectUnmapMemory(): connect %p, memoryType %p, fromTask %p, atAddress %p, returning %i",
          (void *) connect, (void *) memoryType, (void *) fromTask, (void *) atAddress, retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static kern_return_t Hooked_IOServiceOpen(io_service_t service,
                                          task_port_t owningTask,
                                          uint32_t type,
                                          io_connect_t *connect)
{
  kern_return_t retval = IOServiceOpen(service, owningTask, type, connect);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOServiceOpen(): service %p, owningTask %p, type %p, connect %p, returning %i",
          (void *) service, (void *) owningTask, (void *) type,
          (void *) (connect ? *connect : nil), retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static kern_return_t Hooked_IOServiceClose(io_connect_t connect)
{
  kern_return_t retval = IOServiceClose(connect);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOServiceClose(): connect %p, returning %i",
          (void *) connect, retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static kern_return_t Hooked_IOConnectGetService(io_connect_t connect,
                                                io_service_t *service)
{
  kern_return_t retval = IOConnectGetService(connect, service);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOConnectGetService(): connect %p, service %p, returning %i",
          (void *) connect, (void *) (service ? *service : NULL), retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static IONotificationPortRef Hooked_IONotificationPortCreate(mach_port_t masterPort)
{
  IONotificationPortRef retval = IONotificationPortCreate(masterPort);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IONotificationPortCreate(): masterPort %p, returning %p",
          (void *) masterPort, (void *) retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static kern_return_t Hooked_IOConnectSetNotificationPort(io_connect_t connect,
                                                         uint32_t type,
                                                         mach_port_t port,
                                                         uintptr_t reference)
{
  kern_return_t retval = IOConnectSetNotificationPort(connect, type, port, reference);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOConnectSetNotificationPort(): connect %p, type %p, port %p, reference %p, returning %i",
          (void *) connect, (void *) type, (void *) port, (void *) reference, retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static kern_return_t Hooked_IOCreateReceivePort(uint32_t msgType, mach_port_t *recvPort)
{
  kern_return_t retval = IOCreateReceivePort(msgType, recvPort);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOCreateReceivePort(): msgType %p, recvPort %p, returning %i",
          (void *) msgType, (void *) (recvPort ? *recvPort : nil), retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static kern_return_t Hooked_IOConnectAddRef(io_connect_t connect)
{
  kern_return_t retval = IOConnectAddRef(connect);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOConnectAddRef(): connect %p (userRetainCount %u, kernelRetainCount %u), returning %i",
          (void *) connect,
          IOObjectGetUserRetainCount(connect), IOObjectGetKernelRetainCount(connect),
          retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static kern_return_t Hooked_IOConnectRelease(io_connect_t connect)
{
  kern_return_t retval = IOConnectRelease(connect);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOConnectRelease(): connect %p (userRetainCount %u, kernelRetainCount %u), returning %i",
          (void *) connect,
          IOObjectGetUserRetainCount(connect), IOObjectGetKernelRetainCount(connect),
          retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static kern_return_t Hooked_IOObjectRetain(io_object_t object)
{
  kern_return_t retval = IOObjectRetain(object);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOObjectRetain(): object %p, returning %i",
          (void *) object, retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static kern_return_t Hooked_IOConnectSetCFProperties(io_connect_t connect,
                                                     CFTypeRef properties)
{
  kern_return_t retval = IOConnectSetCFProperties(connect, properties);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOConnectSetCFProperties(): connect %p, returning %i, properties:",
          (void *) connect, retval);
    CFShow(properties);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static kern_return_t Hooked_IOConnectSetCFProperty(io_connect_t connect,
                                                   CFStringRef propertyName,
                                                   CFTypeRef property)
{
  kern_return_t retval = IOConnectSetCFProperty(connect, propertyName, property);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOConnectSetCFProperty(): connect %p, propertyName %@, returning %i, property:",
          (void *) connect, propertyName, retval);
    CFShow(property);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static kern_return_t Hooked_IOConnectAddClient(io_connect_t connect,
                                               io_connect_t client)
{
  kern_return_t retval = IOConnectAddClient(connect, client);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"IOConnectAddClient(): connect %p, client %p, returning %i",
          (void *) connect, (void *) client, retval);
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static char *Hooked_getenv(const char *name)
{
  char *retval = getenv(name);
  if (IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    NSLog(@"getenv(): name %s, returning %s\n",
          name ? name : "null", retval ? retval : "null");
    PrintStackTrace();
    OSAtomicDecrement32(&sLogging);
  }
  return retval;
}

static kern_return_t Hooked_mach_port_allocate(ipc_space_t task,
                                               mach_port_right_t right,
                                               mach_port_name_t *name)
{
  static int32_t entered = 0;
  OSAtomicIncrement32(&entered);
  kern_return_t retval = mach_port_allocate(task, right, name);
  if ((entered == 1) && IsMainThread() && IsFromGeForceDriver()) {
    OSAtomicIncrement32(&sLogging);
    //nsAutoreleasePool localPool;
    printf("mach_port_allocate(): task %p, right %p, name %p\n",
           (void *) task, (void *) right,
           (void *) (name ? *name : nil));
    OSAtomicDecrement32(&sLogging);
  }
  OSAtomicDecrement32(&entered);
  return retval;
}

//static CFMachPortRef Hooked_CGWindowServerCFMachPort()
//{
//  CFMachPortRef retval = CGWindowServerCFMachPort();
//  OSAtomicIncrement32(&sLogging);
//  nsAutoreleasePool localPool;
//  NSLog(@"CGWindowServerCFMachPort(): mach_port_t %p, returning %p",
//        (void *) (retval ? CFMachPortGetPort(retval) : 0), retval);
//  PrintStackTrace();
//  OSAtomicDecrement32(&sLogging);
//  return retval;
//}

#pragma mark -

struct interpose_substitution {
  const void* replacement;
  const void* original;
};

#define INTERPOSE_FUNCTION(function) \
    { reinterpret_cast<const void*>(Hooked_##function), \
      reinterpret_cast<const void*>(function) }

__attribute__((used)) static const interpose_substitution substitutions[]
    __attribute__((section("__DATA, __interpose"))) = {
  INTERPOSE_FUNCTION(pthread_once),
  INTERPOSE_FUNCTION(NSPushAutoreleasePool),
  INTERPOSE_FUNCTION(clhDeviceGet),
  INTERPOSE_FUNCTION(clhDeviceDestroy),
  INTERPOSE_FUNCTION(IOConnectMapMemory),
  INTERPOSE_FUNCTION(IOConnectUnmapMemory),
  INTERPOSE_FUNCTION(IOConnectCallMethod),
  INTERPOSE_FUNCTION(IOServiceOpen),
  INTERPOSE_FUNCTION(IOServiceClose),
  INTERPOSE_FUNCTION(IOConnectGetService),
  INTERPOSE_FUNCTION(IONotificationPortCreate),
  INTERPOSE_FUNCTION(IOConnectSetNotificationPort),
  INTERPOSE_FUNCTION(IOCreateReceivePort),
  INTERPOSE_FUNCTION(IOConnectAddRef),
  INTERPOSE_FUNCTION(IOConnectRelease),
  INTERPOSE_FUNCTION(IOObjectRetain),
  INTERPOSE_FUNCTION(IOConnectSetCFProperties),
  INTERPOSE_FUNCTION(IOConnectSetCFProperty),
  INTERPOSE_FUNCTION(IOConnectAddClient),
  INTERPOSE_FUNCTION(getenv),
  INTERPOSE_FUNCTION(mach_port_allocate),
  //INTERPOSE_FUNCTION(CGWindowServerCFMachPort),
};

// What follows are definitions of the Symbolication interfaces we use to get
// stack traces.  The Symbolication framework is an undocumented, private
// framework available on OS X 10.5 and up.  It's used by Apple utilities
// such as dtrace.  http://www.cocoadev.com/index.pl?StackTraces has basic
// information on how to use the Symbolication framework.

@interface VMUAddressRange : NSObject <NSCoding>
{
}

@end

@interface VMUSymbol : VMUAddressRange <NSCopying>
{
}

- (NSString *)name;
- (NSString *)description;
- (id)text;

@end

@interface VMUSymbolOwner : NSObject <NSCopying>
{
}

- (NSString *)name;

@end

// We must override the debugMapExtractorWithMachOHeader method:  Otherwise
// the Symbolication framework can take a minute or more to initialize!!
// We don't need the information in the "debug maps" (Dwarf or dsym debugging
// information) to get simple stack traces.
@interface NSObject (VMUDebugMapExtractorMethodSwizzling)
+ (id)VMUDebugMapExtractor_debugMapExtractorWithMachOHeader:(id)machHeader;
@end

@implementation NSObject (VMUDebugMapExtractorMethodSwizzling)

+ (id)VMUDebugMapExtractor_debugMapExtractorWithMachOHeader:(id)machHeader
{
  return nil;
}

@end

@interface VMUSymbolicator : NSObject
{
}

+ (VMUSymbolicator *)symbolicatorForPid:(int)pid;
- (VMUSymbol *)symbolForAddress:(unsigned long long)address;
- (VMUSymbolOwner *)symbolOwnerForAddress:(unsigned long long)address;

@end

@interface VMUSymbolicator (PerformanceOptimization)
- (id)symbolOwnerNameForAddress:(unsigned long long)fp8;
@end

char *GetOwnerName(void *address)
{
  static char holder[1024] = {0};
  if (!loadSymbolication()) {
    holder[0] = 0;
    return holder;
  }
  VMUSymbolicator *symbolicator = [gSymbolicatorClass symbolicatorForPid:getpid()];
  if (!symbolicator) {
    holder[0] = 0;
    return holder;
  }

  VMUSymbol *symbol =
    [symbolicator symbolForAddress:(unsigned long long)address];
  NSString *ownerName = nil;
  if ([symbolicator respondsToSelector:@selector(symbolOwnerNameForAddress:)]) {
    ownerName =
      [symbolicator symbolOwnerNameForAddress:(unsigned long long)address];
  } else {
    VMUSymbolOwner *owner =
      [symbolicator symbolOwnerForAddress:(unsigned long long)address];
    ownerName = [owner name];
  }
  if (ownerName) {
    strcpy(holder, [ownerName UTF8String]);
    return holder;
  }
  holder[0] = 0;
  return holder;
}

char *GetAddressName(void *address)
{
  static char holder[1024] = {0};
  if (!loadSymbolication()) {
    holder[0] = 0;
    return holder;
  }
  VMUSymbolicator *symbolicator = [gSymbolicatorClass symbolicatorForPid:getpid()];
  if (!symbolicator) {
    holder[0] = 0;
    return holder;
  }

  VMUSymbol *symbol =
    [symbolicator symbolForAddress:(unsigned long long)address];
  if (symbol && [symbol name]) {
    strcpy(holder, [[symbol name] UTF8String]);
    return holder;
  }
  holder[0] = 0;
  return holder;
}

void PrintAddress(void *address)
{
  if (!loadSymbolication()) {
    return;
  }
  VMUSymbolicator *symbolicator = [gSymbolicatorClass symbolicatorForPid:getpid()];
  if (!symbolicator) {
    return;
  }

  VMUSymbol *symbol =
    [symbolicator symbolForAddress:(unsigned long long)address];
  NSString *ownerName = nil;
  if ([symbolicator respondsToSelector:@selector(symbolOwnerNameForAddress:)]) {
    ownerName =
      [symbolicator symbolOwnerNameForAddress:(unsigned long long)address];
  } else {
    VMUSymbolOwner *owner =
      [symbolicator symbolOwnerForAddress:(unsigned long long)address];
    ownerName = [owner name];
  }
  NSLog(@"    (%@) %@, %p", ownerName, [symbol name], address);
}

void PrintStackTrace(NSArray *addresses)
{
  if (!loadSymbolication()) {
    return;
  }
  VMUSymbolicator *symbolicator = [gSymbolicatorClass symbolicatorForPid:getpid()];
  if (!symbolicator) {
    return;
  }

  if (!addresses) {
    addresses = [NSThread callStackReturnAddresses];
  }
  NSUInteger count = [addresses count];
  for (NSUInteger i = 0; i < count; ++i) {
    NSNumber *item = (NSNumber *) [addresses objectAtIndex:i];
    VMUSymbol *symbol =
      [symbolicator symbolForAddress:[item unsignedLongLongValue]];
    NSString *ownerName = nil;
    if ([symbolicator respondsToSelector:@selector(symbolOwnerNameForAddress:)]) {
      ownerName =
        [symbolicator symbolOwnerNameForAddress:[item unsignedLongLongValue]];
    } else {
      VMUSymbolOwner *owner =
        [symbolicator symbolOwnerForAddress:[item unsignedLongLongValue]];
      ownerName = [owner name];
    }
    NSLog(@"    (%@) %@, %p",
          ownerName, [symbol name], (void *)[item unsignedLongLongValue]);
  }
}

BOOL SwizzleMethods(Class aClass, SEL orgMethod, SEL posedMethod, BOOL classMethods)
{
  Method original = nil;
  Method posed = nil;

  if (classMethods) {
    original = class_getClassMethod(aClass, orgMethod);
    posed = class_getClassMethod(aClass, posedMethod);
  } else {
    original = class_getInstanceMethod(aClass, orgMethod);
    posed = class_getInstanceMethod(aClass, posedMethod);
  }

  if (!original || !posed)
    return NO;

  method_exchangeImplementations(original, posed);

  return YES;
}
